#include "general/config.h"

#include <list>

#include "general/mutexlock.hh"
#include "general/TaskThread.hh"
#include "general/ThreadPool.hh"

typedef	std::list< General::TaskThread* >	thread_container_type;


//-----------------------------------------------------------------------
/// \brief Obtain lock
///
/// The use of a function ensures the proper initialization of the data
/// without having to depend on the link order initialization.
//-----------------------------------------------------------------------
inline MutexLock::lock_type&
baton( )
{
  static MutexLock::lock_type	baton =
    MutexLock::Initialize( );

  return baton;
}

//-----------------------------------------------------------------------
/// \brief Obtain the queue of available resources
///
/// The use of a function ensures the proper initialization of the data
/// without having to depend on the link order initialization.
//-----------------------------------------------------------------------
inline static thread_container_type&
available( )
{
  static thread_container_type retval;

  return retval;
}

//-----------------------------------------------------------------------
/// \brief Obtain the queue of resources currently in use.
///
/// The use of a function ensures the proper initialization of the data
/// without having to depend on the link order initialization.
//-----------------------------------------------------------------------
inline static thread_container_type&
in_use( )
{
  static thread_container_type retval;

  return retval;
}

namespace General
{
  //---------------------------------------------------------------------
  /// Secure a resource from the pool of resouces.
  /// If there is no resouce available at the time of the call,
  /// a new resouce is allocated and returned to the caller.
  //---------------------------------------------------------------------
  TaskThread* ThreadPool::
  Acquire( )
  {
    MutexLock	lock( baton( ) );

    if ( available( ).size( ) == 0 )
    {
      available( ).push_front( new TaskThread );
    }
    in_use( ).push_front( available( ).front( ) );
    available( ).pop_front( );
    return in_use( ).front( );
  }

  //---------------------------------------------------------------------
  /// Recyle the resource.
  //---------------------------------------------------------------------
  void ThreadPool::
  Relinquish( TaskThread* Resource )
  {
    if ( Resource )
    {
      MutexLock lock( baton( ) );

      in_use( ).remove( Resource );
      available( ).push_back( Resource );
    }
  }
  
  //---------------------------------------------------------------------
  /// \note
  ///     This method should not be called by others.
  ///     It is provided to help facilitate cleanup at program exit.
  //---------------------------------------------------------------------
  void ThreadPool::
  Reset( )
  {
    MutexLock	lock( baton( ) );
    //-------------------------------------------------------------------
    /// Get rid of threads currently doing something interesting
    //-------------------------------------------------------------------
    for ( thread_container_type::const_iterator
	    cur = in_use( ).begin( ),
	    last = in_use( ).end( );
	  cur != last;
	  ++cur )
    {
      delete( *cur );
    }
    in_use( ).erase( in_use( ).begin( ), in_use( ).end( ) );
    //-------------------------------------------------------------------
    /// Get rid of threads waiting to do something interesting
    //-------------------------------------------------------------------
    for ( thread_container_type::const_iterator
	    cur = available( ).begin( ),
	    last = available( ).end( );
	  cur != last;
	  ++cur )
    {
      delete( *cur );
    }
    available( ).erase( available( ).begin( ), available( ).end( ) );
  }

}
